/*---------------------------------------------------------------------------------
 bbEdgeFlip ( 2004 Slade Taylor [bladestaylor@yahoo.com])
 ----------------------------------------------------------------------------------
 bbEdgeFlip is a plugin for Blackbox for Windows.  For more information,
 please visit [http://bb4win.org] or [http://sourceforge.net/projects/bb4win].
 ----------------------------------------------------------------------------------
 bbEdgeFlip is free software, released under the GNU General Public License,
 version 2, as published by the Free Software Foundation.  It is distributed
 WITHOUT ANY WARRANTY--without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  Please see the GNU General Public
 License for more details:  [http://www.fsf.org/licenses/gpl.html].
 --------------------------------------------------------------------------------*/

#include "bbEdgeFlip.h"

//---------------------------------------------------------------------------------

int beginPlugin(HINSTANCE hPluginInstance)
{
    WNDCLASS wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.lpfnWndProc = PluginProc;
    wc.hInstance = hPluginInstance;
    wc.lpszClassName = pluginInfo(APPNAME);
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    RegisterClass(&wc);

    hFlipWnd = CreateWindowEx(
        WS_EX_TOOLWINDOW,
        pluginInfo(APPNAME),
        NULL,
        WS_POPUP,
        0, 0, 0, 0,
        NULL,
        NULL,
        hPluginInstance,
        NULL);

    SendMessage(hCoreWnd = GetBBWnd(), BB_REGISTERMESSAGE, (WPARAM)hFlipWnd, (LPARAM)bb_messages);

    *(FARPROC*)&pGetDesktopInfo =
        GetProcAddress(GetModuleHandle("blackbox.exe"), "GetDesktopInfo");  // if xoblite

    int i = GetModuleFileName(hPluginInstance, RCpath, MAX_LINE_LENGTH);

    char hook_path[MAX_LINE_LENGTH];
    strcpy(hook_path, RCpath);

    if (i > 3) { RCpath[i - 3] = 'r'; RCpath[i - 2] = 'c'; RCpath[i - 1] = 0; }

    // --- load WMHook.dll ------------------------------
    strcpy(hook_path + i - 14, "WMHook.dll");
    if (hWinMoveHook = LoadLibrary(hook_path))
    {
        *(FARPROC*)&pSetHook = GetProcAddress(hWinMoveHook, "SetHook");
        *(FARPROC*)&pUnsetHook = GetProcAddress(hWinMoveHook, "UnsetHook");
        if (pSetHook && pUnsetHook)
            pSetHook();
    }
    // --------------------------------------------------

    Reconfig();
    return 0;
}

//---------------------------------------------------------------------------------

void endPlugin(HINSTANCE hPluginInstance)
{
    DestroyWindow(hFlipWnd);
    SendMessage(hCoreWnd, BB_UNREGISTERMESSAGE, (WPARAM)hFlipWnd, (LPARAM)bb_messages);
    UnregisterClass(pluginInfo(APPNAME), hPluginInstance);

    if (pUnsetHook)
        pUnsetHook();

    if (hWinMoveHook)
        FreeLibrary(hWinMoveHook);
}

//---------------------------------------------------------------------------------

LRESULT CALLBACK PluginProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        //-------------------------------------------

        case FLIP_WM_MOUSEMOVE:
            {
                POINT pt;
                pt.x = (long)(short)lParam;
                pt.y = (long)(short)wParam;
                if (pt.x > RightEdgeThres)
                {
                    if (FlipControl == CONTROL_NULL)
                    {
                        FlipControl = CONTROL_RIGHT;
                        goto set_tmr;
                    }
                }
                else if (pt.x < LeftEdgeThres)
                {
                    if (FlipControl == CONTROL_NULL)
                    {
                        FlipControl = CONTROL_LEFT;
set_tmr:                UINT interval = 50;
                        if (bTriggerType == TRIGGER_SLIDE)
                            InitialMousePoint = pt;
                        else
                            interval = ((GetAsyncKeyState(VK_CONTROL) & 0x8000) != 0) ? 0: InitialDelay;
                        SetTimer(hwnd, FLIP_TIMER, interval, NULL);
                        SetTimer(hwnd, DCHECK_TIMER, 50, NULL);
                    }
                }
                else
                    FlipControl = CONTROL_NULL;
            }
            break;

        //-------------------------------------------

        case FLIP_WM_MOVING:
            GetCursorPos(&MouseEventPoint);
            if (MouseEventPoint.x < (VScreenX + 3))
                ProcessWindowMove((HWND)wParam, 6, VScreenX + VScreenWidth - 1);
            else if (MouseEventPoint.x > (VScreenX + VScreenWidth - 3))
                ProcessWindowMove((HWND)wParam, 7, VScreenX);
            else
                bControlMove = false;
            break;

        //-------------------------------------------

        case WM_TIMER:
            GetCursorPos(&MouseEventPoint);
            switch (wParam)
            {
                case DCHECK_TIMER:
                    if ((MouseEventPoint.x > RightEdgeThres) ||
                        (MouseEventPoint.x < LeftEdgeThres)) break;
                    goto kill_tm;

                case FLIP_TIMER:
                    if (ProcessFlip(
                        CONTROL_RIGHT,
                        DESK_RIGHT,
                        MouseEventPoint.x > RightEdgeThres,
                        MouseEventPoint.x < LeftEdgeThres)) break;
                    if (ProcessFlip(
                        CONTROL_LEFT,
                        DESK_LEFT,
                        MouseEventPoint.x < LeftEdgeThres,
                        MouseEventPoint.x > RightEdgeThres)) break;
kill_tm:            KillTimer(hwnd, DCHECK_TIMER);
                    KillTimer(hwnd, FLIP_TIMER);
                    FlipControl = CONTROL_NULL;
            }
            break;

        //-------------------------------------------

        case BB_RECONFIGURE:
            Reconfig();
            break;

        //-------------------------------------------

        default:
            return DefWindowProc(hwnd, message, wParam, lParam);

        //-------------------------------------------
    }
    return 0;
}

//---------------------------------------------------------------------------------

bool ProcessFlip(FLIP_CONTROL flip_test, WPARAM wParam, bool test_before, bool test_after)
{
    int new_x;
    if (FlipControl == flip_test)
    {
        if (test_before)
        {
            if (bTriggerType == TRIGGER_SLIDE)
            {
                long test = (MouseEventPoint.y - InitialMousePoint.y);
                if (((test < 0) ? -test: test) > SlideThres)
                {
                    InitialMousePoint.y = MouseEventPoint.y;
                    if (bActionType == ACTION_WINDOWS)
                    {
                        InitialMousePoint.x = MouseEventPoint.x;
                        goto set_2;
                    }
                    else //...if (bActionType == ACTION_UNIX)
                    {
                        InitialMousePoint.x =
                            new_x = VScreenX + ((FlipControl == CONTROL_RIGHT) ? 50: (VScreenWidth - 50));
                        goto set_1;
                    }
                }
            }
            else if (bActionType == ACTION_WINDOWS)
            {
                SetTimer(hFlipWnd, FLIP_TIMER, RepeatDelay, NULL);
                goto set_2;
            }
            else //...if (bActionType == ACTION_UNIX)
            {
                new_x = VScreenX + ((FlipControl == CONTROL_RIGHT) ? 0: VScreenWidth);
                SetTimer(hFlipWnd, FLIP_TIMER, 50, NULL);
set_1:          SetCursorPos(new_x, MouseEventPoint.y);
set_2:          PostMessage(hCoreWnd, BB_WORKSPACE, wParam, NULL);
            }
            return true;
        }
        return test_after;
    }
    return false;
}

//---------------------------------------------------------------------------------

void ProcessWindowMove(HWND h, WPARAM wp, int new_x)
{
    if (!bControlMove)
    {
        if (pGetDesktopInfo)    // 0.0.90/bbLean
        {
            bool b = CheckSticky(h);
            if (!b) MakeSticky(h);
            TransferWindow(wp, h, new_x);
            pGetDesktopInfo(&CurrentDesktopInfo);
            SetTaskWorkspace(h, CurrentDesktopInfo.number);
            if (!b) RemoveSticky(h);
        }
        else                    // xoblite
            TransferWindow(wp, h, new_x);
        bControlMove = true;
    }
}

//---------------------------------------------------------------------------------

void TransferWindow(WPARAM wp, HWND h, int new_x)
{
    SendMessage(hCoreWnd, BB_WORKSPACE, wp, (LPARAM)h);
    SetCursorPos(new_x, MouseEventPoint.y);
}

//---------------------------------------------------------------------------------

void Reconfig()
{
    RCSettings();

    int screen_width = GetSystemMetrics(SM_CXSCREEN);
    int screen_height = GetSystemMetrics(SM_CYSCREEN);

    VScreenHeight = GetSystemMetrics(SM_CYVIRTUALSCREEN);
    VScreenWidth = GetSystemMetrics(SM_CXVIRTUALSCREEN);

    if ((VScreenWidth > screen_width) || (VScreenHeight > screen_height))
        VScreenX = GetSystemMetrics(SM_XVIRTUALSCREEN);
    else
    {
        VScreenX = 0;
        VScreenWidth  = screen_width;
        VScreenHeight = screen_height;
    }

    LeftEdgeThres = VScreenX + EdgeWidth;
    RightEdgeThres = VScreenX + VScreenWidth - EdgeWidth - 1;
    SlideThres = SlideThresPercent * VScreenHeight / 100;
}

//---------------------------------------------------------------------------------

void RCSettings()
{
    char ActionText[8], TriggerText[6];

    RC_Int(     "Width:",               &EdgeWidth,                 1);
    RC_String(  "Action:",              ActionText,                 "Unix",     7);
    RC_String(  "Trigger:",             TriggerText,                "Delay",    5);
    RC_Int(     "Delay.Initial:",       &InitialDelay,              250);
    RC_Int(     "Delay.Repeat:",        &RepeatDelay,               500);
    RC_Int(     "Slide.Distance:",      &SlideThresPercent,         30);

    bActionType = (0 != _stricmp(ActionText, "Windows"));
    bTriggerType = (0 != _stricmp(TriggerText, "Slide"));
}

//---------------------------------------------------------------------------------

void RC_Int(char *label, int *target, int default_val)
{
    *target = ReadInt(RCpath, label, default_val);
    WriteInt(RCpath, label, *target);
}

//---------------------------------------------------------------------------------

void RC_String(char *label, char *target, char *default_val, int max_len)
{
    strncpy(target, ReadString(RCpath, label, default_val), max_len);
    target[max_len] = 0;
    WriteString(RCpath, label, target);
}

//---------------------------------------------------------------------------------
